#include <exception>
#include <map>
#include <string>
#include <inttypes.h>
#include "cda_compiler_support.h"
#include "Ifacexpcom.hxx"
#include "xpcom.hh"

#ifdef IN_GLUE_MODULE
#define PUBLIC_CORBA_GLUE CDA_EXPORT
#else
#define PUBLIC_CORBA_GLUE CDA_IMPORT
#endif

class CCIFactory 
{
public:
  CCIFactory() {}
  /* Mainly just to keep the compiler happy. */
  virtual ~CCIFactory() {}

  /* Find the name of the interface(namespace::interface) */
  virtual const char* Name() const = 0;
  virtual iface::XPCOM::IObject* MakeCCI(XPCOM::IObject_ptr aObj,
                                         PortableServer::POA_ptr aPp)
    const = 0;
} PUBLIC_CORBA_GLUE;

class SCIFactory
{
public:
  SCIFactory() {}
  /* Mainly just to keep the compiler happy. */
  virtual ~SCIFactory() {}

  /* Find the name of the interface(namespace::interface) */
  virtual const char* Name() const = 0;
  virtual XPCOM::IObject_ptr MakeSCI(iface::XPCOM::IObject* aObj,
                                     PortableServer::POA_ptr aPp) const = 0;
} PUBLIC_CORBA_GLUE;

class WrapperRepository
{
public:
  void RegisterCCIFactory(CCIFactory* aCCI) PUBLIC_CORBA_GLUE;
  void RegisterSCIFactory(SCIFactory* aSCI) PUBLIC_CORBA_GLUE;
  iface::XPCOM::IObject* NewCCI(const char* interfaceName,
                                XPCOM::IObject_ptr aObj,
                                PortableServer::POA_ptr aPp) PUBLIC_CORBA_GLUE;
  XPCOM::IObject_ptr NewSCI(const char* interfaceName,
                            iface::XPCOM::IObject* aObj,
                            PortableServer::POA_ptr aPp) PUBLIC_CORBA_GLUE;

private:
  std::map<std::string, CCIFactory*> mCCIFactories;
  std::map<std::string, SCIFactory*> mSCIFactories;
};

WrapperRepository& gWrapperRepository() PUBLIC_CORBA_GLUE;
